(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of CCRecMin.
 * 
 * CCRecMin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CCRecMin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with CCRecMin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

Require Import vector.
Require Import recursive.
Require Import minski_ops.
Set Implicit Arguments.



Section Max.

  (* maximum sur les entiers *)

  Definition max(x y: nat) := match le_lt_dec x y with left _ => y | right _ => x end.

  Lemma le_max_l: forall x y, x<=max x y. Proof. intros x y; unfold max; destruct (le_lt_dec x y); auto. Qed.
  Lemma le_max_r: forall x y, y<=max x y. Proof. intros x y; unfold max; destruct (le_lt_dec x y); auto with arith. Qed.

  Lemma le_max_l': forall x x' y, x<=x' -> x<=max x' y. 
  Proof. 
    intros x x' y xx'; unfold max; destruct (le_lt_dec x' y); auto.
    apply le_trans with x'; assumption.
  Qed.
  Lemma le_max_r': forall x y y', y<=y' -> y<=max x y'. 
  Proof. 
    intros x y y' yy'; unfold max; destruct (le_lt_dec x y'); auto. 
    apply le_trans with y'; auto with arith.
  Qed.

  Lemma minus_mSn_n: forall m n, m+S n-n = S m. 
  Proof. intros; rewrite plus_comm; rewrite plus_Sn_m; rewrite plus_n_Sm; auto with arith. Qed.

  Lemma minus_Sn_n: forall n, S n-n = 1. Proof minus_mSn_n 0. 

End Max.



Section Constant.

  (* Machine constante : on doit utiliser autant d'tats que la constante *)

  Definition Constant(n: nat): MinskiS :=
    [ ! (fix f (n: nat) := 
          match n with 
          | 0 => { Clear (lt_O_Sn _) }
          | S n => { Concat (! f n) (Incr _ 1 0 (lt_O_Sn _) 1 (le_n _) :: nnil _) }
          end) n ].
  
  Theorem ConstantSem: forall n k (x: vector nat k), MinskiSem (Constant n) x n.
  Proof.
    intros n k x; unfold Constant.
    apply SemR; intro x0.
    exists (change (overwrite x x0) 0 n); split.
    induction n; simpl.
    apply ClearSemR; reflexivity.
    MinskiSemR.
    apply IHn.
    step_transition; vauto; apply transition_refl.
    vauto.
  Qed.

End Constant.



Section Zero.

  (* Machine nulle *)

  Definition Zero: MinskiS := [Clear (lt_O_Sn 0)].
  
  Theorem ZeroSem: forall k (x: vector nat k), MinskiSem Zero x 0.
  Proof.
    intros k x; unfold Zero.
    apply SemR; intro x0.
    exists (change (overwrite x x0) 0 0); split.
    apply ClearSemR; reflexivity.
    vauto.
  Qed.

End Zero.



Section Succ.

  (* Machine successeur *)

  Definition Succ: MinskiS := [ Incr 1 1 0 (lt_O_Sn _) 1 (le_n _) :: nnil _ ].
  
  Theorem SuccSem: forall k (x: vector nat (S k)), MinskiSem Succ x (S (head x)).
  Proof.
    intros k x.
    unfold Succ.
    apply SemR; intro x0.
    exists (change (overwrite x x0) 0 (S (head x))); split.
    step_transition; vauto.
    rewrite (get_overwrite_1 (lt_O_Sn k)).
    rewrite <- head_get.
    apply transition_refl.
    vauto.
  Qed.

End Succ.



Section Pik.

  (* Machine projection *)

  Variable i: nat.

  Definition Pik: MinskiS :=
    match eq_nat_dec i 0 with
    | left _ => [ Nil 0 ]
    | right h => [ Move (lt_n_Sn i) (lt_O_Sn i) ]
    end.

  Theorem PikSem: forall k (x: vector nat k) (h: i<k), MinskiSem Pik x (get x i h).
  Proof.
    intros k x h.
    unfold Pik; destruct (eq_nat_dec i 0) as [ik|ik];
    apply SemR; intro x0.

    exists (overwrite x x0); split.
    apply NilSemR; reflexivity.
    subst; rewrite head_get; rewrite (get_overwrite_1 h); reflexivity.

    exists (change (change (overwrite x x0) 0 (get x i h)) i 0); split.
    apply MoveSemR.
    rewrite (get_overwrite_1 h); reflexivity.
    rewrite head_get; get_change; reflexivity.
  Qed.

  Lemma PikSem': forall k (x: vector nat k) y (h: i<k), y=get x i h -> MinskiSem Pik x y.
  Proof. intros; subst; apply PikSem. Qed.

End Pik.



Section Compose.

  (* Compose de la machine F, avec les machines (G i) *)

  Variable i: nat.
  Variable k: nat.
  Variable F: MinskiS.
  Variable G: forall j, j<k -> MinskiS.

  Fixpoint vmax(s: nat): s<S k -> nat := 
    match s return s<S k -> nat with 
    | 0 => fun _ => 0 
    | S s => fun h => max (ns (G (lt_S_n _ _ h))) (vmax (lt_S _ _ (lt_S_n _ _ h)))
    end.

  Let n  := max i (vmax (lt_n_Sn k)).
  Let n' := max k (ns F).

  Let k_n':      k<=n'. Proof. unfold n'. apply le_max_l. Qed.
  Let nsF_n': ns F<=n'. Proof. unfold n'. apply le_max_r. Qed.
  Let i_n:       i<=n.  Proof. unfold n.  apply le_max_l. Qed.

  Let nsGj_n: forall j (h: j<k), ns (G h) <= n.
  Proof.
    cut (forall s (sk: s<S k) j (h: j<k), j<s -> ns (G h) <= vmax sk).
    intros H j h; unfold n; apply le_max_r'; apply H; auto.
    induction s; intros sk j jk js; simpl.
    elim (lt_n_O _ js).
    destruct (le_lt_dec s j) as [sj|sj].
    generalize (le_antisym _ _ (le_S_n _ _ js) sj); intro; subst.
    rewrite (lt_pirr (lt_S_n s k sk) jk); apply le_max_l.
    apply le_max_r'; apply IHs; exact sj.
  Qed.


  Let Sn_SnSn: S n < S n+S n. 
  Proof. pattern (S n) at 1; rewrite (plus_n_O (S n)); apply plus_lt_compat_l; auto with arith. Qed.

  Let n'_nSnSn': n' <= n+S n+ S n'. 
  Proof. auto with arith. Qed.

  Let wG (j: nat) (h: j<k): MinskiM n := { Widen (#G h) (nsGj_n h) }.
  Let wF : MinskiM n' := { Widen (#F) nsF_n'}.
  Let swF : MinskiM (n+S n+S n') := { Shift (!wF) n'_nSnSn'}.


  Let Move_Sn_SnSnj j (h: j<k): Minski 3 (n+S n+S n').
    refine (fun j h => Move (k:=S n) (k':=S n+S n+j) _ _); auto with arith.
    do 2 rewrite <- plus_Sn_m; apply plus_lt_compat_l; apply lt_trans with k; auto with arith.
  Defined.

  Let Move_SnSn_O : Minski 3 (n+S n+S n').
    refine (Move (k:=S n+S n) (k':=0) _ _); auto with arith.
    rewrite (plus_n_O (S n+S n)); do 2 rewrite <- plus_Sn_m; apply plus_lt_compat_l; apply lt_O_Sn.
  Defined.

  Let Gj (j: nat) (h: j<k): MinskiM (n+S n+S n') :=
    {Concat (Widen (MultipleCopy n (S n)) (le_plus_l _ _))
    (Concat (Widen (Shift (!wG h) (le_plus_l n (S n))) (le_plus_l (n+S n) (S n')))
            (Move_Sn_SnSnj h))
    }.

  Fixpoint AllGj (j: nat): j<S k -> MinskiM (n+S n+S n') :=
    match j return j<S k -> MinskiM (n+S n+S n') with
    | 0 => fun h => { Nil _ }
    | S j => fun h => { Concat (!AllGj (lt_S _ _ (lt_S_n _ _ h))) (!Gj (lt_S_n _ _ h)) }
    end.

  Definition Compose: MinskiS := 
     [ Concat (!AllGj (lt_n_Sn k)) (Concat (!swF) Move_SnSn_O) ].


  Section Sem.

    Variable x: vector nat i.
    Variable gx: vector nat k.
    Variable y: nat.

    Hypothesis HG: forall j (h: j<k), MinskiSem (G h) x (get gx j h).
    Hypothesis HF: MinskiSem F gx y.

    Let mcopy_n_Sn: forall (x01 x02: vector nat (S n)) (x03: vector nat (S n')),
      mcopy 
        (truncate (S n + S n) (overwrite x x01 @ x02 @ x03)
          (le_n_S _ _ (le_plus_l _ _)))

        (truncate (S n + S n) (overwrite x x01 @ change (overwrite x x01) n 0 @ x03)
          (le_n_S _ _ (le_plus_l _ _))).

    Proof.
      intros; vauto.
      split; [idtac|split]; intros; vauto.
      do 2 (rewrite (get_append_1 H); vauto).
      rewrite (get_append_1 (lt_S _ _ H)).
      rewrite (get_append_2 (le_plus_r i0 (S n))).
      rewrite get_change_ij.
      apply get_pirr'; rewrite plus_comm; auto with arith.
      rewrite plus_comm; rewrite minus_plus; 
      intro FF; rewrite FF in H; elim (lt_irrefl _ H).
      rewrite (get_append_2 (le_plus_r n (S n)));
      rewrite (get_pirr' (lt_minus (S n) (le_plus_r n (S n)) h) (lt_n_Sn _)).
      apply get_change_ii.
      rewrite plus_comm; auto with arith.
    Qed.


    Let HwG: forall j (h: j<k), MinskiSem [!wG h] x (get gx j h).
    Proof.
      intros j h; simpl; apply WidenSem.
      replace [#G h] with (G h).
      apply HG.
      destruct (G h) as [nn GG]; destruct GG; reflexivity.
    Qed.

    Let wGSemR: forall j (h: j<k) (x0: vector nat (S n)), exists r: vector nat (S n),
      MinskiSemR [!wG h] (overwrite x x0) r /\ get r 0 (lt_O_Sn _) = get gx j h.
    Proof.
      intros j h x0.
      ElimMinskiSem (HwG h x0) r0 H1 H2.
      exists r0; split.
      apply H1.
      rewrite H2; reflexivity.
    Qed.

    Lemma AllGjSem: forall k' (hk': k'<S k) (x0: vector nat (S n+S n+S n')), 
      exists rG: vector nat (S n), forall (r': vector nat (S n+S n+S n')),
      r' = (overwrite x (split1 _ _ (split1 _ _ x0))) @ rG @ (overwrite (truncate k' gx (lt_n_Sm_le _ _ hk')) (split2 _ _ x0)) -> 
      MinskiSemR [!AllGj hk'] (overwrite x x0) r'.
    Proof.
      induction k'; intros hk' x0;
      pose (x01 := split1 _ _ (split1 _ _ x0));
      pose (x02 := split2 _ _ (split1 _ _ x0));
      pose (x03 := split2 _ _ x0);
      pose (xx01:= overwrite x x01);
      pose (x01':= change x01 n 0);
      fold x01; fold xx01; fold x03.

      exists x02; intros r' Hr'.
      simpl; apply NilSemR.
      rewrite Hr'; simpl.
      apply (split_eq (overwrite x x0)); [ apply split_eq; vauto | vauto ]. 
      do 2 rewrite split1_overwrite'; reflexivity.
      rewrite split1_overwrite'.
      unfold x02; apply vector_eq; intros j hj; vauto; rewrite get_overwrite_2; vauto.
      unfold x03; apply vector_eq; intros j hj; vauto; rewrite get_overwrite_2; vauto.

      elim (IHk' (lt_S _ _ (lt_S_n _ _ hk')) x0); clear IHk'; intros rGk' HrGk'.
      elim (wGSemR (lt_S_n _ _ hk') x01'); intros rGk HrGk; elim HrGk; clear HG HrGk; intros HrGk HrGk0.
      exists (change rGk 0 0); intros r' Hr'.
      Opaque MultipleCopy wG.
      simpl; unfold Move_Sn_SnSnj.
      MinskiSemR; (MiscSemR; try apply mcopy_n_Sn; auto) || auto.

      apply (eq_from_append (n:=S n') (xx01 @ rGk')).

      rewrite (change_overwrite_2 i_n). 
      rewrite (truncate_append (n:=S n') (xx01 @ overwrite x x01'));
      rewrite (etruncate_append xx01 (overwrite x x01')).
      cut (MinskiSemR [!wG (lt_S_n k' k hk')] (overwrite x x01')
        (etruncate (S n) (truncate (S (n + S n)) 
        (xx01 @ rGk @ overwrite (truncate k' gx (lt_n_Sm_le k' k (lt_S k' k (lt_S_n k' k hk')))) (split2 (S n + S n) (S n') x0))
        (le_n_S (n + S n) (n + S n + S n') (le_plus_l (n + S n) (S n')))) (le_n_S n (n + S n) (le_plus_l n (S n))))).
      intro HT; apply HT.
      rewrite truncate_append; rewrite (etruncate_append xx01 rGk).
      apply HrGk.

      rewrite minus_plus.
      rewrite (truncate_append (n:=S n') (xx01@change xx01 n 0)).
      rewrite (truncate_append (n:=S n') (xx01@rGk)).
      intros j hj hj' Snj; rewrite (get_append_1 Snj xx01 rGk); 
      apply (get_append_1 Snj xx01 (change xx01 n 0)).

      apply (eq_from_append (A:=nat) (m:=S n+S n) (n:=S n')).

      rewrite Hr'.
      apply (split_eq (m:=S n+S n) (n:=S n')); 
      [ apply split_eq; do 2 (rewrite (split1_change (S n+S n) (S n'))); vauto
      | vauto ].
       rewrite <- plus_assoc; vauto;
       pattern (S n) at 3; rewrite (plus_n_O (S n)); vauto.

       pattern (S n) at 9; rewrite (plus_n_O (S n)); vauto.

       rewrite (split2_change' (m:=S n+S n) (S n')); try exact Sn_SnSn;
       rewrite (split2_change (S n+S n) (S n'));
       rewrite split2_append; fold x03.
       apply vector_eq; intros j hj.
       lt_eq_lt j k' jk'.
        rewrite get_change_ij.
        rewrite (get_overwrite_1 jk'); rewrite (get_overwrite_1 (lt_S _ _ jk')); vauto.
        intro; subst; elim (lt_irrefl _ jk').
        subst; vauto.

        rewrite (get_overwrite_1 (lt_n_Sn k')).
        rewrite (get_append_1 Sn_SnSn (xx01 @ rGk) (m:=S n')). 
        rewrite (get_append_2 (le_n (S n))).
        rewrite (get_pirr' (lt_minus (S n) (le_n (S n)) Sn_SnSn) (lt_O_Sn n)); auto with arith.
        rewrite HrGk0; vauto.

        rewrite get_change_ij.
        do 2 (rewrite get_overwrite_2; vauto).
        intro; subst; elim (lt_irrefl _ jk').
    Qed.

    Let HwF: MinskiSem [!wF] gx y.
    Proof.
      simpl; apply WidenSem.
      replace [#F] with F.
      apply HF.
      destruct F as [nn FF]; destruct FF; reflexivity.
    Qed.

    Let wFSemR: forall (x0: vector nat (S n')), exists r: vector nat (S n'),
      MinskiSemR [!wF] (overwrite gx x0) r /\ get r 0 (lt_O_Sn _) = y.
    Proof.
      intro x0.
      ElimMinskiSem (HwF x0) r0 H1 H2.
      exists r0; split.
      apply H1.
      reflexivity.
    Qed.

    Let swFSemR: forall (x0: vector nat (S n')), exists r: vector nat (S n'),
      get r O (lt_O_Sn _) = y /\
      forall a: vector nat (S n+S n), MinskiSemR [!swF] (a @ overwrite gx x0) (a @ r).
    Proof.
      intro x0.
      elim (wFSemR x0); intros r0 Hr0; elim Hr0; clear Hr0; intros Hr0 Hr0'.
      exists r0; split; [ exact Hr0' | intro a ].
      unfold swF; rewrite Minskimm'; apply ShiftSemR.
      rewrite (etruncate_append a (overwrite gx x0)).
      rewrite (etruncate_append a r0).
      exact Hr0.
      replace (n + S n + S n' - n') with (S n+S n).
      intros j hj hj' jSnSn;
      rewrite (get_append_1 jSnSn a r0);
      apply (get_append_1 jSnSn a (overwrite gx x0)).
      rewrite <- (plus_n_Sm (n+S n)); rewrite <- plus_Sn_m;
      rewrite <- (plus_comm n'); rewrite minus_plus; reflexivity.
    Qed.

    Theorem ComposeSem: MinskiSem Compose x y.
    Proof.
      unfold Compose; apply SemR.
      change (S (n+S n+S n')) with (S n+S n+S n');
      intro x0.
      pose (x01 := split1 _ _ (split1 _ _ x0));
      pose (x02 := split2 _ _ (split1 _ _ x0));
      pose (x03 := split2 _ _ x0);
      pose (xx01 := overwrite x x01).

      elim (AllGjSem (lt_n_Sn k) x0); intros rG0 HGs.
      elim (swFSemR x03); intros xF HxF; elim HxF; clear HxF; intros HxF' HxF.

      exists (change xx01 0 y @ rG0 @ change xF 0 0); split.
      Opaque swF.
      unfold Move_SnSn_O.
      MinskiSemR; auto.
        rewrite truncate_id; apply HxF.
        apply MoveSemR.
        apply (split_eq (m:=S n+S n) (n:=S n')); 
        [ apply (split_eq (m:=S n) (n:=S n));
          do 2 (rewrite (split1_change (S n+S n) (S n'))); vauto
        | vauto ].
          rewrite (get_append_2 (le_n _) (xx01 @ rG0) xF).
          rewrite <- (get_pirr' (lt_O_Sn n') (j:=S n + S n - (S n + S n))).
          rewrite HxF'; reflexivity.
          apply minus_n_n.

          rewrite split2_change'; auto with arith.
          pattern (S n) at 12; rewrite (plus_n_O (S n)); vauto.

          pattern (S n+S n) at 14; rewrite (plus_n_O (S n+S n)).
          rewrite (split2_change (S n+S n) (S n'));
          rewrite (split2_change' (S n') (xx01 @ rG0 @ xF)); vauto.

      rewrite head_get;
      rewrite (get_append_1 (lt_O_Sn (n+S n)) (change xx01 0 y @ rG0) (change xF 0 0));
      rewrite (get_append_1 (lt_O_Sn n) (change xx01 0 y) rG0);
      vauto.
    Qed.

  End Sem.

End Compose.



Section Prim.

  (* Rcursion primitive sur les machines F et G *)

  Variable k: nat.
  Variable F: MinskiS.
  Variable G: MinskiS.

  Let m := ms G.
  Let n := max (ns G) (S (S (S k))).

  Let SSSk_n:    S(S(S k))<= n.     Proof le_max_r _ _.
  Let SSk_n:     S(S k)<= n.        Proof. auto with arith. Qed.
  Let O_n:       0<n.               Proof. apply lt_le_trans with (S (S k)); auto with arith. Qed.
  Let O_SnSSn:   0<S n+S (S n).     Proof. auto with arith. Qed.
  Let SO_SnSSn:  1<S n+S (S n).     Proof. apply (lt_n_S 0 (n+S (S n))); rewrite plus_comm; auto with arith. Qed.
  Let SSO_SnSSn: 2<S n+S (S n).     Proof. apply (lt_n_S 1 (n+S (S n))); rewrite plus_comm; auto with arith. Qed.
  Let n_SnSSn:   n<S n+S (S n).     Proof. auto with arith. Qed.
  Let Sn_SnSSn:  S n<S n+S (S n).   Proof. pattern (S n) at 1; rewrite (plus_n_O (S n)); auto with arith. Qed.
  Let nSn_nSSn:  n+S n<=n+S (S n).  Proof. auto with arith. Qed.
  Let SSn_SnSSn: S (S n)<S n+S (S n). 
  Proof. rewrite <- plus_n_Sm; rewrite plus_Sn_m; do 2 apply lt_n_S; auto with arith. Qed.

  Section PrimIter.

    Let wG: Minski m n           := Widen (#G) (le_max_l _ _).
    Let sG: Minski m (n+S (S n)) := Shift  wG  (le_plus_l n (S (S n))).

    Definition PrimIter: MinskiS :=
      [Loop 0 
          (Concat (Widen (MultipleCopy n (S n)) nSn_nSSn)
          (Concat  sG
          (Concat (Move SSn_SnSSn SSO_SnSSn)
                  (Inkr SO_SnSSn))))
       O_SnSSn].

    Section Sem.

      Variable y e: nat.
      Variable x: vector nat k.

      Lemma PrimIterSemRO: forall x0: vector nat (S n+S (S n)), get x0 0 O_SnSSn = 0 -> MinskiSemR PrimIter x0 x0.
      Proof. intros x0 H; unfold PrimIter; apply LoopSemO; auto. Qed.

      Variables z t h h': nat.
      Hypothesis HG: MinskiSem G (t::h::x) h'.
      Hypothesis HR: forall x0: vector nat (S n+S (S n)), 
        exists x0': vector nat (S n+S (S n)),
        get x0' 2 SSO_SnSSn = y /\ MinskiSemR PrimIter (overwrite (z::S t::h'::x) x0) x0'.

      Let HwG: MinskiSem [wG] (t::h::x) h'.
      Proof.
        unfold wG; apply WidenSem.
        replace [#G] with G.
        apply HG.
        clear HR; destruct G as [nn GG]; destruct GG; reflexivity.
      Qed.

      Let sGSemR: forall (x0: vector nat (S n)), exists r: vector nat (S n),
        get r 0 (lt_O_Sn n) = h' /\
        forall (a: vector nat (S n)) b,
        MinskiSemR [sG] (a @ b :: overwrite (t::h::x) x0) (a @ b :: r).
      Proof.
        intro x0.
        ElimMinskiSem (HwG x0) r0 H1 H2.
        exists r0; split. 
        vauto.
        intros a b; unfold sG; apply ShiftSemR.
        do 2 rewrite (etruncate_append_cons (n:=S n) a).
        apply H1.
        rewrite minus_plus.
        intros j hj hj' jSn.
        lt_eq_lt (S n) j Snj.
        elim (lt_irrefl _ (in_int_lt _ _ _ (in_int_intro _ _ _ Snj jSn))).
        subst.
        rewrite (get_append_2 (le_n (S n)) a (b::r0)).
        rewrite (get_append_2 (le_n (S n)) a (b::overwrite (t::h::x) x0)).
        do 2 (rewrite <- (get_pirr' (j:=S n-S n) (lt_O_Sn (S n))); try rewrite <- minus_n_n; vauto).
        rewrite (get_append_1 Snj a (b::r0));
        apply (get_append_1 Snj a (b::overwrite (t::h::x) x0)).
      Qed.

      Let mcopy_n_Sn: forall x0: vector nat (S n+S (S n)),
        mcopy (truncate (S n+S n) (overwrite (z::t::h::x) x0) (le_n_S (n + S n) (n + S (S n)) nSn_nSSn))
              (truncate (S n+S n) 
                 (split1 (S n) (S (S n)) (overwrite (z::t::h::x) x0) @
                 overwrite (change (split1 (S n) (S (S n)) (overwrite (z::t::h::x) x0)) n 0) (split2 _ _ x0))
                 (le_n_S (n + S n) (n + S (S n)) nSn_nSSn)).
      Proof.
        intro x0; split; [idtac | split]; intros; vauto.
        rewrite (get_append_1 H).
        destruct (le_lt_dec (S (S (S k))) i) as [SSSki|SSSki].
        do 2 (rewrite (get_overwrite_2 SSSki); vauto).
        do 2 (rewrite (get_overwrite_1 SSSki); vauto).
        rewrite (get_append_2 (le_plus_r i (S n))).
        rewrite <- (get_pirr' (j:=i+S n-S n) (lt_S _ _ (lt_S _ _ H))).
        rewrite split1_overwrite'.
        rewrite change_overwrite_2; vauto.
        rewrite (get_overwrite_1 (lt_S _ _ H)).
        destruct (le_lt_dec (S (S (S k))) i) as [SSSki|SSSki].
        do 2 rewrite (get_overwrite_2 SSSki); rewrite get_change_ij; vauto; intro FF; rewrite FF in H; elim (lt_irrefl _ H).
        do 2 rewrite (get_overwrite_1 SSSki); vauto.
        rewrite plus_comm; rewrite minus_plus; reflexivity.
        rewrite (get_append_2 (le_plus_r n _) (split1 (S n) (S (S n)) (overwrite (z :: t :: h :: x) x0))).
        rewrite <- (get_pirr' (j:=n+S n-S n) (lt_S _ _ (lt_n_Sn n))).
        rewrite (get_overwrite_1 (lt_n_Sn n)); vauto.
        rewrite plus_comm; rewrite minus_plus; reflexivity.
      Qed.

      Lemma PrimIterSemRS: forall x0: vector nat (S n+S (S n)), exists x0': vector nat (S n+S (S n)),
        get x0' 2 SSO_SnSSn = y /\ MinskiSemR PrimIter (overwrite (S z::t::h::x) x0) x0'.
      Proof.
        intro x0; unfold PrimIter.

        elim (sGSemR (tail (overwrite (change (split1 (S n) (S (S n)) x0) n 0) (split2 _ _ x0)))); intros rG HrG; 
        elim HrG; clear HrG sGSemR; intros HrG HrG'.
        elim (HR (split1 (S n) (S (S n)) x0 @ z :: change rG 0 0)).
        intros r0 Hr0; elim Hr0; clear Hr0 HR HG; intros Hr0 Hr0'.

        exists r0; split.
        exact Hr0.
        eapply LoopSemS; auto.
          rewrite (get_overwrite_1 (lt_O_Sn (S (S k)))); unfold get; reflexivity.
          MinskiSemR; (MiscSemR; auto) || auto.
            rewrite (change_overwrite_1 (lt_O_Sn (S (S k)))); unfold change.
            apply mcopy_n_Sn.
            rewrite split1_overwrite'; rewrite (change_overwrite_2 SSSk_n).
            rewrite (change_overwrite_1 (lt_O_Sn (S (S k)))).
            intros j hj hj' jSnSn.
            cut (j=S n+S n). intro; subst.
            rewrite get_overwrite_2.
            rewrite (get_append_2' (S n) hj').
            rewrite get_overwrite_2; vauto.
            apply le_trans with n; auto with arith.
            rewrite <- plus_n_Sm in hj; apply le_antisym; auto with arith.

            rewrite split1_overwrite'; rewrite (change_overwrite_2 SSSk_n).
            replace (overwrite (overwrite (z :: t :: h :: x) (change (split1 (S n) (S (S n)) x0) n 0)) (split2 (S n) (S (S n)) x0))
              with (z::overwrite (t :: h :: x) (tail (overwrite (change (split1 (S n) (S (S n)) x0) n 0) (split2 (S n) (S (S n)) x0)))).
            apply HrG'.
            apply vector_eq; intros j hj; destruct j.
            rewrite (get_overwrite_1 (lt_O_Sn n)); rewrite (get_overwrite_1 (lt_O_Sn (S (S k)))); unfold get; reflexivity.
            change (get (z::overwrite (t::h::x) (tail (overwrite (change (split1 (S n) (S (S n)) x0) n 0) (split2 (S n) (S (S n)) x0)))) (S j) hj)
              with (get (overwrite (t::h::x) (tail (overwrite (change (split1 (S n) (S (S n)) x0) n 0) (split2 (S n) (S (S n)) x0)))) j (lt_S_n _ _ hj)).
            destruct (le_lt_dec (S n) (S j)) as [hj'|hj'].            
            rewrite (get_overwrite_2 hj');  vauto;
            rewrite get_overwrite_2; try apply le_trans with n; auto with arith;
            rewrite get_tail; rewrite (get_overwrite_2 hj'); vauto.
            rewrite (get_overwrite_1 hj').
            destruct (le_lt_dec (S (S k)) j) as [SSkj|SSkj].
            do 2 (rewrite get_overwrite_2; vauto).
            rewrite get_tail; rewrite (get_overwrite_1 hj'); vauto.
            rewrite (get_overwrite_1 SSkj); rewrite (get_overwrite_1 (lt_n_S _ _ SSkj)). 
            change (get (z::t::h::x) (S j) (lt_n_S _ _ SSkj)) with (get (t::h::x) j (lt_S_n _ _ (lt_n_S _ _ SSkj))); vauto.
            
            fold PrimIter.
            rewrite (change_append_1 (A:=nat) (m:=S n) (n:=(S (S n)))).
            rewrite change_overwrite_1; auto with arith.
            rewrite (get_append_2 (le_n_Sn (S n)) (overwrite (z :: t :: h :: x) (split1 (S n) (S (S n)) x0)) (z::rG)).
            rewrite <- (get_pirr' (j:=S (S n)-S n) (lt_n_S _ _ (lt_O_Sn n))).
            rewrite get_change_ij; auto with arith.
            pose (a:= change (z :: t :: h :: x) 2 (get (z :: rG) 1 (lt_n_S 0 (S n) (lt_O_Sn n)))); fold a.
            rewrite (get_append_1 (lt_n_S _ _ O_n) (overwrite a (split1 (S n) (S (S n)) x0)) (z::rG)).
            rewrite (get_overwrite_1 (lt_n_S _ _ (lt_O_Sn (S k)))).
            pattern (S (S n)) at 5; replace (S (S n)) with (S n+1); auto with arith.
            rewrite (change_append_2 (overwrite a (split1 (S n) (S (S n)) x0)) (z :: rG)).
            rewrite (change_append_1 (A:=nat) (m:=S n) (n:=S (S n))); auto with arith.
            rewrite change_overwrite_1. 
            unfold a.
            rewrite get_cons; rewrite HrG.
            unfold change at 1 2 3, get.
            change (change (z :: rG) 1 0) with (z::change rG 0 0).
            rewrite <- (overwrite_append_1 (le_n_S _ _ SSk_n)). 
            apply Hr0'.
            auto with arith.
            rewrite plus_comm; reflexivity.
            rewrite minus_Sn_n; reflexivity.
            apply le_lt_trans with (S (S k)); auto with arith.
      Qed.

    End Sem.

  End PrimIter.

  Let sF: MinskiS := Compose (S k) F (fun j (h: j<k) => Pik (S j)).
    
  Definition Prim: MinskiS := 
    Compose (S k)
      [Concat (#PrimIter) (Move SSO_SnSSn O_SnSSn)]
      (fun j (h: j<S (S (S k))) => 
        match j with
        | 0 => Pik 0
        | 1 => Zero
        | 2 => sF
        | (S (S j)) => Pik j
        end).

  Section Sem.

    Variable y : nat.
    Variable x : vector nat (S k).
    Variable SemF: vector nat k -> nat -> Prop.
    Variable SemG: vector nat (S (S k)) -> nat -> Prop.

    Hypothesis HSemF: Injective SemF.
    Hypothesis HSemG: Injective SemG.
    Hypothesis HF:    forall x y, SemF x y -> MinskiSem F x y.
    Hypothesis HG:    forall x y, SemG x y -> MinskiSem G x y.
    Hypothesis HIter: IterSem SemF SemG (head x) (tail x) y.

    Let PrimIterSemR: forall z t h (x0: vector nat (S n+S (S n))), 
      IterSem SemF SemG (t+z) (tail x) y ->
      IterSem SemF SemG  t    (tail x) h ->
      exists r': vector nat (S n+S (S n)), 
        get r' 2 SSO_SnSSn = y /\ MinskiSemR PrimIter (overwrite (z :: t :: h :: tail x) x0) r'.
    Proof.
      induction z; intros t h x0 Htz_y Ht_h.
        exists (overwrite (0 :: t :: h :: tail x) x0); split.
        rewrite (get_overwrite_1 (lt_n_S _ _ (lt_n_S _ _ (lt_O_Sn k)))); simpl.
        rewrite <- plus_n_O in Htz_y.
        apply (IterSemInjective HSemF HSemG t (tail x) h y); assumption.
        apply PrimIterSemRO; rewrite (get_overwrite_1 (lt_O_Sn (S (S k)))); reflexivity.

        rewrite <- plus_n_Sm in Htz_y. 
        elim (IterSemReverse _ _ _ (S t) z y Htz_y); intros gth Hgth; elim Hgth; clear Hgth; intros h2 Hgth2.
        elim Hgth2; clear Hgth2; intros Hgth2 Hgth.
        generalize (IterSemInjective HSemF HSemG t (tail x) h2 h Hgth2 Ht_h); 
        intro h2h; rewrite h2h in Hgth; clear Hgth2 h2h h2.
        apply PrimIterSemRS with gth; auto.
        intro x1; apply IHz; auto.
        exists h; auto.
    Qed.

    Theorem PrimSem: MinskiSem Prim x y.
      rewrite (head_tail x). 
      elim (IterSemReverse _ _ _ 0 (head x) y HIter); intros yf1 Hf1.
      unfold Prim; apply ComposeSem with (head x :: 0 :: yf1 :: tail x).
      intros j hj; destruct j; [idtac | destruct j; [idtac | destruct j]].
        apply PikSem' with (lt_O_Sn k); reflexivity.
        apply ZeroSem.
        simpl; unfold sF; apply ComposeSem with (tail x).
          intros j' hj'; apply PikSem' with (lt_n_S _ _ hj'); simpl; vauto.
          exact (HF Hf1).
        apply PikSem' with (lt_S_n _ _ (lt_S_n _ _ hj)); reflexivity.
        apply SemR; intro x0.
        elim (PrimIterSemR (head x) 0 yf1 x0); auto. 
        intros r' Hr'; elim Hr'; clear Hr'; intros Hr'y Hr'.
        exists (change (change r' 0 y) (S (S O)) 0); split.
        MinskiSemR; (MiscSemR; auto) || auto.
          apply Hr'.
          rewrite <- Hr'y; reflexivity.
          rewrite head_get; get_change; reflexivity.
    Qed.

  End Sem.

End Prim.



Section Min.

  (* Minimisation sur la machine F *)

  Variable k: nat.
  Variable F: MinskiS.

  Let n := max (S (ns F)) k.

  Let k_n:      k<= n.           Proof le_max_r _ _.
  Let nsF_Sn:   ns F<=S n.       Proof. apply le_trans with (S (ns F)); [ auto with arith | exact (le_S _ _ (le_max_l _ _)) ]. Qed.
  Let neq_O_n:  0<>n.            Proof. generalize (le_max_l (S (ns F)) k); intros H FF; fold n in H; rewrite <- FF in H; elim (le_Sn_O _ H). Qed.
  Let O_SnSSn:  O<S n+S (S n).   Proof. auto with arith. Qed.
  Let n_SnSSn:  n<S n+S (S n).   Proof. auto with arith. Qed.
  Let Sn_SnSSn: S n<S n+S (S n). Proof. auto with arith. Qed.
  Let Sn_nSSn:  S n<=n+S (S n).  Proof. auto with arith. Qed.
  Let n_nSSn':   n<>n+S (S n).    
  Proof. 
    intro FF; pattern n at 1 in FF; rewrite (plus_n_O n) in FF;
    generalize (plus_reg_l _ _ _ FF); intro HH; discriminate HH.
  Qed.
  Let Sn_nSSn':  S n<>n+S (S n).  
  Proof. 
    intro FF; pattern (S n) at 1 in FF; replace (S n) with (n+1) in FF.
    generalize (plus_reg_l _ _ _ FF); intro HH; discriminate HH.
    rewrite plus_comm; reflexivity.
  Qed.

  Let wF:  MinskiM (S n)       := {Widen (#F)  nsF_Sn}.
  Let swF: MinskiM (n+S (S n)) := {Shift (!wF) Sn_nSSn}.

  Let MinIter: MinskiM (n+S (S n)) :=
    {Concat (Copy n_SnSSn Sn_SnSSn (lt_n_Sn _))
    (Concat (MultipleCopy n (S (S n)))
    (Concat (!swF)
            (Inkr n_SnSSn)))}.

  Definition Min: MinskiS  :=
    [Concat (Clear n_SnSSn)
    (Concat (Inkr Sn_SnSSn)
    (Concat (Loop (S n) (!MinIter) Sn_SnSSn)
    (Concat (Decr n_SnSSn)
            (Move n_SnSSn O_SnSSn))))].

  Section Sem.

    Variable y: nat.
    Variable x: vector nat k.
 
    Hypothesis H0: MinskiSem F (y::x) 0.
    Hypothesis Hm: forall y', y'<y -> exists x', MinskiSem F (y'::x) (S x').

    Section SWF.

      Variables z t: nat.
      Hypothesis HF: MinskiSem F (z::x) t.

      Let HwF: MinskiSem [!wF] (z::x) t.
      Proof.
        simpl; apply WidenSem.
        replace [#F] with F.
        apply HF.
        destruct F as [nn FF]; destruct FF; reflexivity.
      Qed.

      Let wFSemR: forall (x0: vector nat (S (S n))), exists r: vector nat (S (S n)),
        MinskiSemR [!wF] (overwrite (z::x) x0) r /\ get r 0 (lt_O_Sn _) = t.
      Proof.
        intro x0.
        ElimMinskiSem (HwF x0) r0 H1 H2.
        exists r0; split.
        apply H1.
        reflexivity.
      Qed.

      Lemma swFSemR: forall (x0: vector nat (S n)), exists r: vector nat (S (S n)),
        get r O (lt_O_Sn _) = t /\
        forall a: vector nat (S n), MinskiSemR [!swF] (a @ z::overwrite x x0) (a @ r).
      Proof.
        intro x0.
        elim (wFSemR (z::x0)); intros r0 Hr0; elim Hr0; clear Hr0; intros Hr0 Hr0'.
        exists r0; split; [ exact Hr0' | intro a ].
        unfold swF; rewrite Minskimm'; apply ShiftSemR.
        rewrite (etruncate_append a (z::overwrite x x0)).
        rewrite (etruncate_append a r0).
        exact Hr0.
        replace (n + S (S n) - S n) with (S n).
        intros j hj hj' jSn;
        rewrite (get_append_1 jSn a r0);
        apply (get_append_1 jSn a (z::overwrite x x0)).
        rewrite <- plus_n_Sm; rewrite <- plus_Sn_m; rewrite minus_plus; reflexivity.
      Qed.

    End SWF.

    Let mcopy_n_SSn: forall (x1: vector nat (S n)) (x2: vector nat (S (S n))), 
      mcopy (x1 @ x2) (x1 @ (get x2 0 (lt_O_Sn _)) :: (change x1 n 0)).
    Proof.
      intros x1 x2; split; [idtac|split]; intros.
      lt_eq_lt (S n) i Sni.
      elim (lt_irrefl _ (in_int_lt _ _ _ (in_int_intro _ _ _ Sni H))).
      subst; do 2 (rewrite (get_append_2 (le_n (S n))); rewrite <- (get_pirr' (j:=S n-S n) (lt_O_Sn (S n))); vauto).
      do 2 rewrite (get_append_1 Sni); reflexivity.
      cut (S n <= i+S(S n)); [intro Sn_iSSn|auto with arith].
      rewrite (get_append_2 Sn_iSSn).
      rewrite (get_append_1 (lt_S _ _ H)).
      rewrite <- (get_pirr' (j:=i + S (S n) - S n) (lt_n_S _ _ (lt_S _ _ H))).
      simpl; rewrite get_change_ij; vauto.
      intro FF; rewrite FF in H; elim (lt_irrefl _ H).
      rewrite minus_mSn_n; reflexivity.
      rewrite (get_append_2 Sn_nSSn).
      rewrite <- (get_pirr' (j:=n + S (S n) - S n) (lt_n_Sn (S n))). 
      simpl; vauto.
      rewrite minus_mSn_n; reflexivity.
    Qed.


    Let MinIterStep: forall w (x1: vector nat (S n)) (r r': vector nat (S (S n))),
      w<=y -> 
      (forall a : vector nat (S n), MinskiSemR [!swF] (a @ (y-w) :: overwrite x (change x1 n 0)) (a @ r)) -> 
      MinskiSemR [Loop (S n) (!MinIter) Sn_SnSSn]
        (overwrite x (change x1 n (S (y - w))) @ r)
        (overwrite x (change x1 n (S y)) @ r') -> 
      forall x2 p,
      MinskiSemR [Loop (S n) (!MinIter) Sn_SnSSn]
        (change (overwrite x x1) n (y - w) @ change x2 0 (S p))
        (change (overwrite x x1) n (S y) @ r').
    Proof.
      intros w x1 r r' wy Hr HR x2 p.
      eapply LoopSemS.
        rewrite (get_append_2 (le_n (S n)) (change (overwrite x x1) n (y-w)) (change x2 0 (S p))).
        rewrite <- (get_pirr' (j:=S n-S n) (lt_O_Sn (S n))); [ rewrite get_change_ii | rewrite <- minus_n_n ]; reflexivity.
        unfold MinIter; rewrite Minskimm'.
        MinskiSemR; (MiscSemR; auto) || auto; fold (plus n (S (S n))).
          rewrite change_ii_invol.
          rewrite get_change_ij; auto with arith.
          rewrite (get_append_1 (lt_n_Sn n) (change (overwrite x x1) n (y-w)) (change x2 0 (S p))).
          rewrite get_change_ii.
          pattern (S n) at 9; rewrite (plus_n_O (S n)); rewrite (change_append_2 (change (overwrite x x1) n (y-w)) (change x2 0 (S p))).
          pattern (n+S(S n)) at 2; rewrite <- (plus_n_Sm n (S n)); rewrite <- (plus_Sn_m n (S n)).
          rewrite (change_append_2 (change (overwrite x x1) n (y-w)) (change (change x2 0 (S p)) 0 (y-w))).
          apply mcopy_n_SSn.

          rewrite get_change_ij; auto with arith.
          rewrite get_change_ii; rewrite change_ii_invol; do 2 rewrite (change_overwrite_2 k_n).
          apply Hr.

          rewrite (get_append_1 (lt_n_Sn n) (overwrite x (change x1 n (y-w))) r).
          rewrite (get_overwrite_2 k_n); rewrite get_change_ii.
          rewrite (change_append_1 (overwrite x (change x1 n (y-w))) r); auto with arith.
          do 2 rewrite (change_overwrite_2 k_n); rewrite change_ii_invol.
          exact HR.
    Qed.

    Let MinIterSemR: forall (x0: vector nat (S n+S (S n))) p w, 
      w<=y -> 
      exists r: vector nat (S (S n)),
      MinskiSemR [Loop (S n) (!MinIter) Sn_SnSSn] 
        (change (change (overwrite x x0) n (y-w)) (S n) (S p)) 
        (change (overwrite x (split1 _ _ x0)) n (S y) @ r).
    Proof.
      intros x0 p w wy.
      rewrite (split _ _ x0); vauto.
      pose (x1:=split1 _ _ x0); pose (x2:=split2 _ _ x0); fold x1 x2.
      rewrite change_append_1; auto with arith.
      pattern (S n) at 15; rewrite (plus_n_O (S n)); rewrite change_append_2. 

      cut (exists r : vector nat (S (S n)), forall x2 p,
             MinskiSemR [Loop (S n) (!MinIter) Sn_SnSSn]
              (change (overwrite x x1) n (y - w) @ change x2 0 (S p))
              (change (overwrite x x1) n (S y) @ r)).
      intro H; elim H; intros r Hr; exists r; auto.

      gclear x1; clear p x0 x2.
      induction w; intros x1.

        elim (swFSemR H0 (change x1 n 0)); intros r Hr; elim Hr; clear Hr; intros Hr0 Hr.
        exists r; intros x2 p; eapply MinIterStep; try rewrite <- minus_n_O; auto with arith.
        apply LoopSemO; auto.
        rewrite (get_append_2 (le_n _) (overwrite x (change x1 n (S y))) r). 
        rewrite <- (get_pirr' (j:=S n-S n) (lt_O_Sn (S n))); auto with arith.

        cut (y-S w < y); [ intro ySw_y | auto with arith ].
        elim (Hm ySw_y); intros x' Hx'.
        elim (swFSemR Hx' (change x1 n 0)); intros r Hr; elim Hr; clear Hr; intros Hr0 Hr.
        elim IHw with x1; auto with arith; intros r' Hr'.
        exists r'; intros x2 p; eapply MinIterStep; auto with arith.
        do 2 rewrite (change_overwrite_2 k_n) in Hr'.
        replace (S (y-S w)) with (y-w); auto with arith.
        replace r with (change r 0 (S x')).
        apply Hr'.
        apply change_id with (lt_O_Sn (S n)); auto.
        rewrite minus_Sn_m; auto with arith.
    Qed.

    Theorem MinSem: MinskiSem Min x y. 
    Proof.
      unfold Min; apply SemR; fold (S n+S (S n)); intro x0.
      pose (x01 := split1 _ _ x0).
      pose (xx01 := overwrite x x01).
      pose (x02 := tail (split2 _ _ x0)).
      pose (p := get (change (overwrite x x0) n 0) (S n) Sn_SnSSn).
      elim MinIterSemR with x0 p y; auto;
      intros x1 Hx1; rewrite <- minus_n_n in Hx1.
      exists (change (change xx01 0 y) n 0 @ x1); split.
      MinskiSemR; (MiscSemR; auto) || auto.
        apply Hx1.
        fold x01.
        rewrite (change_overwrite_2 k_n); vauto.
        rewrite (get_append_1 (lt_n_Sn n) (overwrite x (change x01 n (S y))) x1).
        rewrite (get_overwrite_2 k_n); vauto; unfold pred.
        rewrite (change_append_1 (overwrite x (change x01 n (S y))) x1); auto.
        rewrite (change_overwrite_2 k_n); rewrite change_ii_invol.
        apply sym_eq; rewrite (change_ij_comm neq_O_n).
        rewrite (change_append_1 (overwrite x (change x01 n y)) x1); auto with arith.
        rewrite (change_overwrite_2 k_n); rewrite change_ii_invol.
        rewrite (change_append_1 (overwrite x (change x01 n 0)) x1); auto with arith.
        unfold xx01; rewrite (change_ij_comm neq_O_n).
        rewrite (change_overwrite_2 k_n); reflexivity.
        rewrite head_get; rewrite (get_append_1 (lt_O_Sn n) (change (change xx01 0 y) n 0) x1); 
        get_change; reflexivity.
    Qed.

  End Sem.

End Min.



Section Compilation.

  (* Compilation des machines rcursives vers les machines de Minski *)

  (* on force la liste pour viter un appel paresseux (et donc dupliqu...) *)
  Let force k (f: forall i, i<k -> MinskiS): forall i, i<k -> MinskiS :=
    get (init k (fun i => match le_lt_dec k i with left _ => Zero | right h => f i h end)).

  Let get_force: forall k (f: forall i, i<k -> MinskiS) i (h: i<k), force f h = f i h.
  Proof.
    intros; unfold force; rewrite get_init.
    destruct (le_lt_dec k i) as [ki|ki].
    elim (lt_irrefl _ (lt_le_trans _ _ _ h ki)).
    rewrite (lt_pirr h ki); reflexivity.
  Qed.


  Fixpoint compile(k: nat)(f: Recfun k) {struct f}: MinskiS :=
    match f with
      | recursive.Constant n      => Constant n 
      | recursive.Zero            => Zero
      | recursive.Succ            => Succ
      | recursive.Pik i _ _       => Pik i
      | recursive.Compose _ i f g => Compose i (compile f) (force (fun j h => compile (g j h)))
      | recursive.Prim k f g      => Prim k (compile f) (compile g)
      | recursive.Min _ f         => Min k (compile f)
    end.


  Theorem CompileCorrect: forall (k: nat) (f: Recfun k) (x: vector nat k) (y: nat),
    RecSem f x y -> MinskiSem (compile f) x y.
  Proof.
    induction f; intros x y Hf; simpl; try rewrite Hf.

      apply ConstantSem.
      apply ZeroSem.
      apply SuccSem.
      apply PikSem; reflexivity.

      (* composition *)
      elim Hf; intros gx Hgx; elim Hgx; clear Hgx; intros Hgx Hgx'.
      apply ComposeSem with gx; auto.
      intros; rewrite get_force; auto. 

      (* recursion *)
      apply PrimSem with (RecSem f1) (RecSem f2); try apply RecSemInjective; auto.
      
      (* minimalization *)
      elim Hf; intros Hf0 Hfm.
      apply MinSem; auto.
      intros y' Hy'; elim (Hfm y' Hy'); intros x' Hx'.
      exists x'; auto.

  Qed.

End Compilation.
