(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of CCRecMin.
 * 
 * CCRecMin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CCRecMin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with CCRecMin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

Require Import vector.
Set Implicit Arguments.

Section Minski.

  Variables m n: nat.

  (* deux types d'instruction possibles *)
  Inductive Step: Set := 
  | Incr (i: nat) (hi: i<S n) (q:    nat) (hq: q<=m)
  | Test (i: nat) (hi: i<S n) (q q': nat) (hq: q<=m) (hq': q'<=m).

  (* machine : liste d'instructions *)
  Definition Minski: Set := vector Step m.

  (* configuration : un tat, sa preuve de validit, et l'tat des registres *)
  Inductive  Config: Set := config (q: nat) (hq: q<=m) (r: vector nat (S n)).

  (* fonctions d'accs *)
  Definition reg  (C: Config) := let (_,_,r) := C in r.
  Definition state(C: Config) := let (q,_,_) := C in q.
  Lemma proof: forall C, state C <= m.
  Proof. destruct C; unfold state; exact hq. Defined.

  (* configuration "arguments" *)
  Definition args(k: nat) (x: vector nat k) (x0: vector nat (S n)): Config :=
    config (le_O_n _) (overwrite x x0).

  (* "rsultat" d'une configuration *)
  Definition result(C: Config) (y: nat): Prop := y = head (reg C).

  (* une configuration est-elle finale ou pas *)
  Definition final (C: Config): Prop := state C=m.
  Definition nfinal(C: Config): Prop := state C<m.
  Definition final_dec(C: Config): {final C} + {nfinal C} :=
    match lt_eq_lt_dec (state C) m with
    | inleft (left h) => right _ h
    | inleft (right h) => left _ h
    | inright h => False_rec _ (lt_irrefl _ (le_lt_trans _ _ _ (proof C) h))
    end.


  Lemma final_nnfinal: forall C, final C -> ~nfinal C.
  Proof. unfold final; unfold nfinal; intros C H H'; rewrite H in H'; elim (lt_irrefl _ H'). Qed.

  Lemma deConfig: forall (C: Config) (q: nat) (r: vector nat (S n)), 
    q=state C -> r=reg C -> forall h, config (q:=q) h r = C.
  Proof. destruct C; intros; subst; rewrite (le_pirr h hq); reflexivity. Qed.

  Lemma reg_eq: forall q h r q' r' h', config (q:=q) h r = config (q:=q') h' r' -> r = r'.
  Proof. intros q h r q' h' r' H; injection H; auto. Qed.

  Lemma state_eq: forall q h r q' r' h', config (q:=q) h r = config (q:=q') h' r' -> q = q'.
  Proof. intros q h r q' h' r' H; injection H; auto. Qed.

End Minski.
Implicit Arguments Incr [].
Implicit Arguments Test [].
Implicit Arguments config [].
Implicit Arguments deConfig [m n].


(* Type des machines incluant leur nombre d'tats *)
Definition MinskiM n: Set := sigS (fun m => Minski m n).
Definition Minskim m n: Minski m n -> MinskiM n := existS (fun m => Minski m n) m.
Definition Minskim' n (M: MinskiM n) := projS2 M. 
Notation "{ m }" := (Minskim m).
Notation "! M" := (Minskim' M) (at level 30).
Ltac MinskiM := unfold Minskim', Minskim, projS1, projS2.
Lemma Minskimm': forall m n (M: Minski m n), Minskim' {M}=M.
Proof. intros; reflexivity. Qed.

(* Type des machines incluant leur nombre d'tats ET de registres *)
Definition MinskiS: Set := sigS MinskiM.
Definition Minskis m n (M: Minski m n): MinskiS := existS MinskiM n { M }.
Definition Minskis' (M: MinskiS) := ! (projS2 M). 
Notation "[ m ]" := (Minskis m).
Notation "# M" := (Minskis' M) (at level 30).
Ltac MinskiS := unfold Minskis', Minskis, projS1, projS2.
Lemma Minskiss': forall m n (M: Minski m n), #[M]=M.
Proof. intros; reflexivity. Qed.


Section Machine.

  Variable MS: MinskiS.

  Definition ms: nat := projS1 (projS2 MS). (* nombre d'tats *)
  Definition ns: nat := projS1 MS.          (* nombre de registres *)
  Definition Ms: Minski ms ns := # MS.      (* machine "dpacke" *)

  (* pas de calul : la configuration doit tre non finale *)
  Definition step(C: Config ms ns) (H: nfinal C): Config ms ns :=
    let r := reg C in
    match get Ms (state C) H with
    | Incr i hi q'    hq'     => config ms ns q' hq' (change r i (S (get r i hi)))
    | Test i hi q1 q2 hq1 hq2 => 
      match get r i hi with
      | 0   => config ms ns q1 hq1 r
      | S x => config ms ns q2 hq2 (change r i x)
      end
    end.
  Implicit Arguments step [].

  Lemma step_pirr:  forall (C: Config ms ns) (H H': nfinal C), step C H = step C H'.
  Proof. intros C H H'; unfold step; rewrite (get_pirr H H'); reflexivity. Defined.


  (* transition : liste de pas de calcul *)
  Inductive transition: Config ms ns -> Config ms ns -> Prop :=
  | transition_refl: forall (C: Config ms ns), transition C C
  | step_transition: forall (C C': Config ms ns) (H: nfinal C), transition (step C H) C' -> transition C C'.

  (* ajout d'un pas  la fin d'une transition *)
  Lemma transition_step: forall (C C': Config ms ns) (H': nfinal C'), transition C C' -> transition C (step C' H').
  Proof.
    intros C C' H' CC'; induction CC' as [|C C' H CC' HR].
    apply step_transition with H'; apply transition_refl.
    apply step_transition with H; apply HR; exact H'.
  Qed.

  (* concatnation de transitions *)
  Lemma transition_transition: forall C D E: Config ms ns, transition C D -> transition D E -> transition C E.
  Proof.
    intros C D E CD; induction CD; intro C'E.
    apply C'E.
    apply step_transition with H; apply IHCD; exact C'E.
  Qed.

  (* les transitions vers les configurations finales sont uniques *)
  Lemma transition_final: forall C C': Config ms ns, transition C C' -> final C -> C=C'.
  Proof.
    intros C C' H; destruct H as [| C C' H CC']; intro Hf.
    reflexivity.
    elim (final_nnfinal Hf H).
  Qed.


  (* formes normales *)
  Definition NF(C C': Config ms ns) := transition C C' /\ final C'.

  (* smantique restreinte (arguments, rsultats) *)
  Definition MinskiSem (k: nat) (x: vector nat k) (y: nat) := 
    forall x0: vector nat (S ns), exists Cf, NF (args ms x x0) Cf /\ result Cf y.
  
  (* smantique complte (relation entre deux tats des registres) *)
  Definition MinskiSemR r r' := transition (config _ _ _ (le_O_n _) r) (config _ _ _ (le_n _) r').


  (* unicit des formes normales *)
  Theorem NF_injective: forall C C1 C2: Config ms ns, NF C C1 -> NF C C2 -> C1=C2.
  Proof.
    intros C C1 C2 H1 H2; 
    destruct H1 as [CC1 C1f];
    destruct H2 as [CC2 C2f].
    gclear4 CC1 CC2 C1f C2f .
    induction 1 as [|C C' H CC' HR].
    intros; apply transition_final; assumption.
    intros; apply HR; try assumption.
    destruct CC2; [ elim (final_nnfinal C2f H) | rewrite (step_pirr H H0); exact CC2 ].
  Qed.

  (* ... et de la smantique *)
  Lemma MinskiSemInjective: forall (k: nat) (x: vector nat k) (y y': nat), 
    MinskiSem x y -> MinskiSem x y' -> y=y'.
  Proof.
    intros k x y y' H H'.
    destruct (H  (fill (S ns) (nnil _) 0)) as [Cf  HCf ]; destruct HCf  as [NFC  Hy ].
    destruct (H' (fill (S ns) (nnil _) 0)) as [Cf' HCf']; destruct HCf' as [NFC' Hy'].
    rewrite (NF_injective NFC NFC') in Hy.
    destruct Cf'; 
    unfold result in Hy;  rewrite Hy;
    unfold result in Hy'; rewrite Hy';
    reflexivity.
  Qed.


  (* configuration fortement normalisables *)
  Inductive SN: Config ms ns -> Prop := 
  | final_SN: forall C, final C -> SN C
  | step_SN : forall C (Hnf: nfinal C), SN (step C Hnf) -> SN C.

  (* ordre sur les configurations SN *)
  Let step_lt(C C': sig SN): Prop := exists h, (proj1_sig C) = step (proj1_sig C') h.
  Infix "<-" := step_lt (at level 60).

  (* bonne fondation de cet ordre *)
  Let step_lt_wf: well_founded step_lt.
  Proof.
    unfold well_founded; destruct a as [C SNC].
    apply (SN_ind (fun C => forall SNC: SN C, Acc step_lt (exist SN C SNC))); 
      assumption || 
      ( clear C SNC; intros;
        apply Acc_intro; intros G SNstepGD;
        destruct G as [G SNG];
        elim SNstepGD; intro nGf ).
    elim (final_nnfinal H nGf).
    intro Heq; unfold proj1_sig in Heq; subst.
    rewrite (step_pirr Hnf nGf) in H0.
    apply H0; exact f.
  Defined.

  Let SN_SN_step: forall (C: Config ms ns) (H: nfinal C), SN C -> SN (step C H).
  Proof.
    intros C Cnf SNC; destruct SNC as [C Cf | C Cnf' HCstep].
    elim (final_nnfinal Cf Cnf).
    rewrite (step_pirr Cnf Cnf'); exact HCstep.
  Qed.
  Let stepSN(C: sig SN) (H: nfinal (proj1_sig C)): sig SN := exist SN (step _ H) (SN_SN_step H (proj2_sig C)).
  Let stepSN_lt: forall (C: sig SN) (H: nfinal (proj1_sig C)), stepSN C H <- C.
  Proof. destruct C; intro H; exists H; reflexivity. Defined.

  (* excution  partir d'une configuration SN *)
  Definition execute(Co: Config ms ns)(SNCo: SN Co): sig (NF Co) :=
    well_founded_induction step_lt_wf (fun (C: sig SN) => transition Co (proj1_sig C) -> sig (NF Co))
      (fun (C: sig SN) (exec: (forall C', C' <- C -> transition Co (proj1_sig C') -> sig (NF Co))) CoC => 
         match final_dec (proj1_sig C) with
         | left  H => exist (NF Co) (proj1_sig C) (conj CoC H)
         | right H => exec (stepSN C H) (stepSN_lt C H) (transition_step H CoC)
         end
      )
      (exist SN Co SNCo)
      (transition_refl Co).
  Implicit Arguments execute [].

  (* excution borne *)
  Definition execute_N N k (x: vector nat k): option nat :=
    (fix e(N: nat): Config ms ns -> option nat :=
      match N with 
      | 0 => fun C => None
      | S N => fun C =>
        match final_dec C with
        | left H => Some (head (reg C))
        | right H => e N (step C H)
        end
      end) N (args _ x (fill _ (nnil _) 0)).

  (* l'existence d'une smantique implique la normalisation forte *)
  Theorem MinskiSem_SN: forall k (x: vector nat k) y, MinskiSem x y -> forall x0, SN (args ms x x0).
  Proof.
    intros k x y H x0; elim (H x0); intros Cf HCf; elim HCf; clear HCf; intros HCf HCf';
    elim HCf; clear HCf; intros HCf HCf''.
    induction HCf; [apply final_SN | eapply step_SN]; auto.
  Qed.

  Lemma transition_step': forall A B C (HB: nfinal B), transition A B -> C = step B HB -> transition A C.
  Proof. intros; subst; apply transition_step; assumption. Qed.

  Lemma step_transition': forall A B C (HA: nfinal A), B = step A HA -> transition B C -> transition A C.
  Proof. intros; subst; apply step_transition with HA; assumption. Qed.

  Lemma transition_eq_transition: forall A B C D, B=C -> transition A B -> transition C D -> transition A D.
  Proof. intros A B C D H; rewrite H; intros AB CD; eapply transition_transition; [ apply AB | apply CD ]. Qed.

  Lemma transition_refl': forall q r h q' r' h', q=q' -> r=r' -> transition (config _ _ q h r) (config _ _ q' h' r'). 
  Proof. intros; subst; rewrite (le_pirr h h'); apply transition_refl. Qed.

End Machine.

(* passage de la smantique complte  la smantique restreinte *)
Lemma SemR: forall m n (M: Minski m n) k (x: vector nat k) y,
  (forall x0: vector nat (S n), exists r': vector nat (S n), MinskiSemR [M] (overwrite x x0) r' /\ y = head r') ->
  MinskiSem [M] x y.
Proof.
  intros m n M k x y H x0.
  elim (H x0); intros r' Hr'; elim Hr'; clear Hr'; intros Hr' Hhr'.
  exists (config m n m (le_n _) r'); split. split.
  unfold args; apply Hr'.
  unfold final; reflexivity.
  rewrite Hhr'; unfold result; reflexivity.
Qed.


(* tactiques *)

Ltac msns := unfold ms, ns, projS1, projS2; simpl.
Ltac msnsH H := unfold ms, ns, projS1, projS2 in H; simpl in H.
Ltac MsNs M m n := change (ms [M]) with m; change (ns [M]) with n.
Ltac MsNsH M m n H := change (ms [M]) with m in H; change (ns [M]) with n in H.

Ltac final := unfold final; unfold state; reflexivity.
Ltac nfinal C :=
  (cut (nfinal C)); [ intro | unfold nfinal; unfold state; auto with arith ].
Ltac step_simpl := unfold step; msns.


Ltac transition_step C := 
  nfinal C; unfold MinskiSemR;
  match goal with H: nfinal C |- transition _ _ _ => 
    apply transition_step' with C H; [ step_simpl | auto ]; 
    try clear H 
  end.


Ltac step_transition := 
  unfold MinskiSemR;
  match goal with |- transition ?M ?C _ => 
    nfinal C; 
    match goal with H: nfinal C |- _ =>
      apply step_transition' with (step M C H) H; [ auto | try step_simpl ];
      try clear H
    end
  end.

Ltac step_transition_tac tac := 
  unfold MinskiSemR;
  match goal with |- transition [?M] ?C _ => 
    nfinal C; 
    match goal with H: nfinal C |- _ =>
      apply step_transition with H; 
      unfold step, state, Ms, M; rewrite Minskiss';
      try rewrite tac; fold M; msns;
      try clear H
    end
  end.

Ltac transition_refl := 
  unfold MinskiSemR;
  match goal with 
  | |- transition ?M (config ?m ?n _ _ _) _ => 
    change m with (ms M); change n with (ns M);
    apply transition_refl'; auto
  | |- transition ?M _ (config ?m ?n _ _ _) => 
    change m with (ms M); change n with (ns M);
    apply transition_refl'; auto
  | |- transition _ _ _ => apply transition_refl'; auto
  end.

Ltac step_SN := 
  match goal with |- SN _ ?C => 
    nfinal C; 
    match goal with H: nfinal C |- _ => 
      apply step_SN with H; step_simpl
    end 
  end.
Ltac final_SN := apply final_SN; final.

Ltac NF tac := split; [ apply tac; auto | final ].

Ltac MinskiSem r := 
  match goal with |- MinskiSem ?MS _ _ => 
    exists (config (ms MS) (ns MS) _ (le_n _) r); split; 
    [ split; 
      [ unfold args; try rewrite overwrite_full; msns 
      | final ] 
    | unfold result, reg; vauto ]
  end.

Ltac ElimMinskiSem H r H1 H2 := 
  elim H; intros C_ HC_; destruct C_ as [q_ hq_ r];
  elim HC_; clear HC_; intros H1 H2; elim H1; clear H1; intros H1 Hf_;
  unfold final, state in Hf_; subst;
  unfold result, reg in H2; rewrite head_get in H2; unfold args in H1;
  generalize (le_pirr hq_ (le_n _)); intro; subst.



(* exemple: l'addition *)
Section MAdd.

  Definition MAdd: MinskiS. 
  refine [
    Test 2 1 1 _ 2 1 _ _ :: 
    Incr 2 1 0 _ 0   (le_O_n _)  ::
    nnil _
  ]; auto with arith. Defined.

  Theorem MAddSem: forall m n: nat, MinskiSem MAdd (m::n::nnil _) (m+n).
  Proof.
    intros m n.
    pose (MAdd':=#MAdd); simpl in MAdd'.
    MinskiSem (m+n::0::nnil _).
    transition_step (config 2 1 0 (le_O_n _) (m + n :: 0 :: nnil nat)).
    gclear m;
    induction n; intro m.
    rewrite <- plus_n_O; apply transition_refl.
    step_transition.
    step_transition.
    rewrite <- plus_n_Sm; rewrite <- plus_Sn_m.
    apply IHn.
  Qed.

  Let MAddSN: forall m n: nat, SN MAdd (config 2 1 0 (le_O_n _) (m::n::nnil _)).
  Proof.
    intros m n; gclear2 n m.
    induction n; intro m; step_SN;
    [ final_SN | step_SN; apply IHn ].
  Defined.

  Definition ExecMAdd(m n: nat): nat := 
    match execute MAdd (config 2 1 0 (le_O_n _) (m::n::nnil _)) (MAddSN m n) with
    | exist (config _ _ r) _ => head r
    end.

  Let huit: ExecMAdd 5 3 = 8.
  Proof. reflexivity. Qed.

End MAdd.
