(*
 * Copyright (C) 2004 Damien Pous
 * 
 * This file is part of CCRecMin.
 * 
 * CCRecMin is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * CCRecMin is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with CCRecMin; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *)

Require Import vector.
Set Implicit Arguments.

(* type inductif des fonctions rcursives *)
Inductive Recfun: nat -> Set :=
| Constant: nat -> Recfun 0
| Zero:     Recfun 1
| Succ:     Recfun 1
| Pik:      forall i k, i < k -> Recfun k
| Compose:  forall k i, Recfun k -> (forall j, j<k -> Recfun i) -> Recfun i
| Prim:     forall k  , Recfun k -> Recfun (S (S k)) -> Recfun (S k)
| Min:      forall k  , Recfun (S k) -> Recfun k.
Implicit Arguments Pik [].

(* injectivit d'une relation *)
Definition Injective k (P: vector nat k -> nat -> Prop): Prop :=
  forall (x: vector nat k) y y', P x y -> P x y' -> y=y'.


(* itration d'une smantique (pour dfinir la smantique de la rcursion) *)
Section IterSem.

  Variable k: nat.
  Variable SemF: vector nat k -> nat -> Prop.
  Variable SemG: vector nat (S (S k)) -> nat -> Prop.

  (* on droule *)
  Fixpoint IterSem (t: nat): vector nat k -> nat -> Prop :=
    match t with
    | 0 => SemF
    | S t => fun x y => exists h, IterSem t x h /\ SemG (t :: h :: x) y
    end.

  (* Lemmes pour la compilation vers les machines de Minski : il faut pouvoir inverser la rcursion... *)
  Lemma IterSemReverse: forall (x: vector nat k) t z yzt, 
    IterSem (t+z) x yzt -> exists yt: nat, IterSem t x yt.
  Proof.
    induction t; induction z; intros yzt H.
      exists yzt; exact H.

      elim H; intros h Hh; elim Hh; clear Hh; intros Hh Hh'.
      apply IHz with h; exact Hh.

      elim H; intros h Hh; elim Hh; clear Hh H; simpl; rewrite <- plus_n_O; intros Hh Hh'.
      exists yzt; exists h; split; assumption.

      elim H; intros h Hh; elim Hh; clear Hh H; simpl; rewrite <- plus_n_Sm; intros Hh Hh'.
      apply IHz with h; exact Hh.
  Qed.

  (* ...et identifier *)
  Lemma IterSemInjective: Injective SemF -> Injective SemG -> forall t, Injective (IterSem t).
  Proof.
    intros HF HG; induction t; simpl; intros x y y' Hy Hy'.
    apply HF with x; auto.
    elim Hy;  intros h  Hh ; elim Hh ; clear Hh ; intros Hh  Hhg.
    elim Hy'; intros h' Hh'; elim Hh'; clear Hh'; intros Hh' Hhg'.
    rewrite IHt with x h h' in Hhg; auto.
    apply HG with (t::h'::x); auto.
  Qed.

End IterSem.


(* smantique des fonction rcursives *)
Fixpoint RecSem k (f: Recfun k) {struct f} : vector nat k -> nat -> Prop :=
match f in Recfun k return vector nat k -> nat -> Prop with
| Constant n      => fun _ m => m = n
| Zero            => fun _ m => m = 0
| Succ            => fun l m => m = S (head l)
| Pik i k h       => fun l m => m = get l i h
| Compose k i f g => fun l m => exists gl, RecSem f gl m /\ forall j (h: j<k), RecSem (g j h) l (get gl j h)
| Prim k f g      => fun l   => IterSem (RecSem f) (RecSem g) (head l) (tail l)
| Min k f         => fun l m => RecSem f (m::l) 0 /\ forall i, i<m -> exists x, RecSem f (i::l) (S x)
end.


(* correction de cette smantique *)
Theorem RecSemInjective: forall k (f: Recfun k), Injective (RecSem f).
Proof.
  induction f as [n | | | i k h | k i f HRf g HRg | k f HRf g HRg | k f HRf ]; intros l x y; simpl; intros Hx Hy;

  (* Constant, Zero, Succ, Pik *)
  try (rewrite Hx; rewrite Hy; reflexivity).

  (* Compose *)
  elim Hx; intros glx Hglx; clear Hx;
  elim Hglx; intros Hglxf Hglxg; clear Hglx.
  elim Hy; intros gly Hgly; clear Hy;
  elim Hgly; intros Hglyf Hglyg; clear Hgly.
  cut (glx=gly).
  intro Heq; rewrite Heq in Hglxf; apply HRf with gly; assumption.
  apply vector_eq; intros j hj; apply (HRg j hj l); [ apply Hglxg | apply Hglyg ].

  (* Prim *)
  gclear4 x y Hx Hy.
  induction (head l) as [ | n HR ]; simpl.
  apply HRf.
  intros x y Hx Hy.
  elim Hx; intros hx Hhx; elim Hhx; intros HRx Hgx; clear Hx Hhx.
  elim Hy; intros hy Hhy; elim Hhy; intros HRy Hgy; clear Hy Hhy.
  rewrite (HR hx hy) in Hgx; auto.
  apply HRg with (n :: hy :: tail l); [apply Hgx | apply Hgy ].

  (* Min *)
  elim Hx; intros HZx HMx; clear Hx.
  elim Hy; intros HZy HMy; clear Hy.
  apply le_antisym; apply not_gt; intros H;
  cut (exists m, 0 = S m); try (intro F; elim F; intros m Fm; discriminate Fm).
  generalize (HMx y H); intro HEm; elim HEm; intros m Hm; clear HEm;
  exists m; apply HRf with (y::l); assumption.
  generalize (HMy x H); intro HEm; elim HEm; intros m Hm; clear HEm;
  exists m; apply HRf with (x::l); assumption.
  
Qed.


(* utilitaires *)
Definition Compose0 i (f: Recfun 0)                  := Compose f (get (nnil (Recfun i))).
Definition Compose1 i (f: Recfun 1) (g:    Recfun i) := Compose f (get (g :: nnil _)).
Definition Compose2 i (f: Recfun 2) (g g': Recfun i) := Compose f (get (g :: g' :: nnil _)).

Definition RecSem1 (f: Recfun 1) a     n := RecSem f (a :: nnil _) n.
Definition RecSem2 (f: Recfun 2) a b   n := RecSem f (a :: b :: nnil _) n.
Definition RecSem3 (f: Recfun 3) a b c n := RecSem f (a :: b :: c :: nnil _) n.


(* tactiques *)
Ltac absurd_lt h  := elim (lt_irrefl _ h) || elim (lt_n_O _ h) || absurd_lt (lt_S_n _ _ h).
Ltac sreflexivity := simpl; reflexivity.

Ltac bounded_prop := match goal with | |- forall (j: nat) (h: j<_), _ => intros j h; ( absurd_lt h || bounded_prop' j h ) end
with bounded_prop' j h := destruct j; [ idtac | absurd_lt h || bounded_prop' j h ].

Ltac composition gx := exists gx; split; [ try sreflexivity | bounded_prop; unfold get; try sreflexivity ].
Ltac composition0      := composition (nnil nat).
Ltac composition1 x    := composition (x :: nnil _).
Ltac composition2 x x' := composition (x :: x' :: nnil _).

Ltac primitive1 f := match goal with 
                     | |- forall m: nat, RecSem _ (m :: nnil nat) _ => 
                       intro m; induction m as [ | m HR ]; 
                       [ try sreflexivity 
                       | exists (f m); split; [ exact HR | try sreflexivity ] ] 
		     | |- forall m: nat, RecSem1 _ _ _ => unfold RecSem1; primitive1 f
                    end. 
Ltac primitive2 f := match goal with 
		     | |- forall m n: nat, RecSem _ (m :: n :: nnil nat) _ => 
                       intro m; induction m as [ | m HR ]; 
                       [ try sreflexivity 
                       | intro n; exists (f m n); split; [ exact (HR n) | try sreflexivity ] ] 
		     | |- forall m n: nat, RecSem2 _ _ _ _ => unfold RecSem2; primitive2 f
                    end. 

Ltac deCompose H x := 
  elim H; intro x; finite_list x;intro Hx; clear H;
  elim Hx; intros Heq Hg; clear H Hx;
  deCompose_ Hg 0 lt_O_Sn; clear Hg
with
deCompose_ Hg n h := generalize (Hg n (h _)); unfold get; intro; try deCompose_ Hg (S n) (fun x => lt_n_S _ _ (h x)).




(* exemples : adddition, racine carre *)

Definition Id:  Recfun 1 := Pik 0 1 (lt_O_Sn _).
Definition P02: Recfun 2 := Pik 0 2 (lt_O_Sn _).
Definition P12: Recfun 2 := Pik 1 2 (lt_n_Sn _).
Definition P03: Recfun 3 := Pik 0 3 (lt_O_Sn _).
Definition P13: Recfun 3 := Pik 1 3 (lt_trans _ _ _ (lt_n_Sn _) (lt_n_Sn _)).
Definition P23: Recfun 3 := Pik 2 3 (lt_n_Sn _).

Section Addition.

  Definition Add: Recfun 2 := Prim Id (Compose1 Succ P13).

  Theorem AddSem: forall m n: nat, RecSem2 Add m n (m+n).
  Proof.
    primitive2 plus.
    composition1 (m+n).
  Qed.

End Addition.

Section Multiplication.

  Definition Mult: Recfun 2 := Prim Zero (Compose2 Add P13 P23).

  Theorem MultSem: forall m n: nat, RecSem2 Mult m n (m*n).
  Proof.
    primitive2 mult.
    composition2 (m*n) n.
    replace (S m * n) with (m*n + n).
    apply AddSem.
    simpl; auto with arith.
  Qed.

End Multiplication.

Section SquareRoot.

  Definition Sqr: Recfun 1 := Compose2 Mult Id Id.
  Lemma SqrSem: forall n: nat, RecSem1 Sqr n (n*n).
  Proof.
    intro n.
    composition2 n n.
    apply MultSem.
  Qed.
    
  Definition null n := match n with 0 => 0 | _ => 1 end.
  Lemma nullS: forall m n, m < n -> null (n-m) = 1.
  Proof.
    double induction m n; simpl; auto with arith.
    intros n l F; absurd_lt F.
  Qed.
  Definition Null: Recfun 1 := Prim (Constant 0) (Compose0 _ (Constant 1)).
  Lemma NullSem: forall n: nat, RecSem1 Null n (null n).
  Proof.
    primitive1 null.
    composition0. 
  Qed.

  Definition null2 m n := null (m+n).
  Definition Null2: Recfun 2 := Compose1 Null Add.
  Lemma Null2Sem: forall m n: nat, RecSem2 Null2 m n (null2 m n).
  Proof.
    intros m n.
    composition1 (m+n).
    unfold null2; apply NullSem.
    apply AddSem.
  Qed.

  Definition Pred: Recfun 1 := Prim (Constant 0) (Compose1 Id P02).
  Lemma PredSem: forall n: nat, RecSem1 Pred n (pred n).
  Proof.
    primitive1 pred.
    composition1 m.
  Qed.

  Definition SubInv: Recfun 2 := Prim Id (Compose1 Pred P13).
  Lemma SubInvSem: forall m n: nat, RecSem2 SubInv m n (n - m).
  Proof.
    primitive2 (fun m n => n - m).
    simpl; auto with arith.
    composition1 (n-m).
    replace (n - S m) with (pred (n - m)).
    apply PredSem.
    generalize m; induction n as [ | n HRn]; auto with arith.
    intro m'; destruct m'; auto with arith.
    simpl; apply HRn.
  Qed.

  Definition Sub: Recfun 2 := Compose2 SubInv P12 P02.
  Lemma SubSem: forall m n: nat, RecSem2 Sub m n (m - n).
  Proof.
    intros m n.
    composition2 n m.
    apply SubInvSem.
  Qed.
 
  Definition beq m n := null2 (m-n) (n-m).
  Definition Eq: Recfun 2 := Compose2 Null2 Sub SubInv.
  Lemma EqSem: forall m n: nat, RecSem2 Eq m n (beq m n).
  Proof.
    intros m n.
    composition2 (m-n) (n-m).
    unfold beq;  apply Null2Sem.
    apply SubSem.
    apply SubInvSem.
  Qed.
  Lemma EqRefl: forall n: nat, RecSem2 Eq n n 0.
  Proof.
    intro n.
    replace 0 with (beq n n).
    apply EqSem.
    unfold beq; rewrite <- minus_n_n; reflexivity.
  Qed.
  Lemma EqNRefl: forall m n: nat, m<>n -> RecSem2 Eq m n 1.
  Proof.
    intros m n H.
    replace 1 with (beq m n).
    apply EqSem.
    unfold beq; unfold null2.
    lt_eq_lt m n H'.
    rewrite not_le_minus_0; simpl; auto with arith; apply nullS; assumption.
    elim H; exact H'.
    rewrite (not_le_minus_0 n m); simpl; auto with arith. rewrite <- plus_n_O; apply nullS; assumption.
  Qed.
  Lemma EqRefl_eq: forall m n: nat, RecSem Eq (m::n::nnil _) 0 -> m=n.
  Proof.
    intros m n H.
    destruct (eq_nat_dec m n) as [ Hmn | Hmn ].
    exact Hmn.
    cut (0=1).
    intro F; discriminate F.
    apply (RecSemInjective Eq (m::n::nnil _)).
    exact H.
    apply EqNRefl; exact Hmn.
  Qed.


  Definition sqrt(m n: nat): Prop := m=n*n. 
  Definition Sqrt: Recfun 1 := Min (Compose2 Eq (Compose1 Id P12) (Compose1 Sqr P02)).

  Lemma sqr_S: forall n: nat, S n*S n = n*n + (n+n + 1).
  Proof.
    intro n.
    rewrite <- mult_n_Sm; rewrite mult_comm; rewrite <- mult_n_Sm.
    rewrite <- (plus_n_Sm (n+n)).
    rewrite <- plus_n_O.
    rewrite plus_n_Sm. 
    rewrite plus_assoc.
    reflexivity.
  Qed.
  Lemma sqr_lt: forall m n: nat, m < n -> m*m < n*n.
  Proof.
    induction m as [ | m HRm ]; intros n H; destruct n.
    absurd_lt H.
    simpl; auto with arith.
    absurd_lt H.
    do 2 rewrite sqr_S.
    apply (plus_lt_le_compat (m*m) (n*n) (m+m+1) (n+n+1)).
    apply (HRm n (lt_S_n _ _ H)).
    do 2 (apply plus_le_compat; auto with arith); apply le_S_n; apply lt_le_weak; exact H.
  Qed.
  Lemma nsqrt: forall m n, sqrt m n -> forall i, i<n -> ~ sqrt m i.
  Proof.
    unfold sqrt; intros m n H i Hi F.
    apply lt_irrefl with m.
    pattern m at 1; rewrite F; rewrite H.
    apply sqr_lt.
    exact Hi.
  Qed.

  Theorem SqrtSem: forall m n: nat, sqrt m n <-> RecSem1 Sqrt m n.
  Proof.
    intros m n; split; intro H.

    split.
    composition2 m (n*n).
    rewrite H; apply EqRefl.
    clear h; composition1 m.
    clear h; composition1 n.
    apply SqrSem.
    intros i Hi.
    exists 0.
    composition2 m (i*i).
    apply EqNRefl; exact (nsqrt H Hi).
    clear h; composition1 m.
    clear h; composition1 i.
    apply SqrSem.

    elim H; intros H0 Hm; clear H Hm.
    deCompose H0 x.
    rewrite <- (EqRefl_eq Heq) in H0; clear Heq.
    deCompose H xm.
    rewrite (RecSemInjective Id (head xm :: _) _ _ (refl_equal (head xm)) Heq) in H; clear xm Heq.
    simpl in H; rewrite H in H0; clear x H.
    deCompose H0 xm.
    rewrite (RecSemInjective P02 (n :: m :: _) _ _ H (refl_equal n)) in Heq; clear xm H.
    unfold sqrt; apply (RecSemInjective Sqr (n::nnil _)); [ exact Heq | apply SqrSem ].
  Qed.

End SquareRoot.
